import sys
import numpy as np
import petsc4py

# bit dirty handling here

usens=False
if "-usens" in sys.argv:
  #print("adding nullspace")
  sys.argv.remove("-usens")
  usens=True

# by default form the PC mat by calculating A^TA so LU works
# if usenormal is passed, fewer PCs will work since only some work with MATNORMAL
usenormal=False
if "-usenormal" in sys.argv:
  sys.argv.remove("-usenormal")
  usenormal=True

# use the system where PCQR broke down
qrdiv=False
if "-qrdiv" in sys.argv:
  sys.argv.remove("-qrdiv")
  qrdiv=True


petsc4py.init(sys.argv)
import petsc4py.PETSc as pe

# qr breakdown
if qrdiv:
    st = 132388
    ddir = 1
else:
    st = 131835
    ddir = 0

bp = "dumps/"

name1 = "matdump_step%08d_dir%d.bin" % (st, ddir)
name2 = "vecdump_step%08d_dir%d.bin" % (st, ddir)
name3 = "vecsoldump_step%08d_dir%d.bin" % (st, ddir)

viewer = pe.Viewer().createBinary(bp+name1)
mymat = pe.Mat()
mymat.load(viewer)
viewer.setFileName(bp+name2)
rhs = pe.Vec()
rhs.load(viewer)

viewer.setFileName(bp+name3)
guess = pe.Vec()
guess.load(viewer)

viewer.destroy()

def nsvec(size):
    # manual creation instead of const=True necessary since mat is rectangular
    arr = np.ones(size)/np.sqrt(size)
    nsvec = pe.Vec().create()
    nsvec.setSizes(size)
    nsvec.setUp()
    nsvec.array = arr
    ns = pe.NullSpace().create(vectors=[nsvec])
    nsvec.destroy()
    return ns

# remove the pinned DoF and add the nullspace
if usens:
    nrow, ncol = mymat.getSize()
    tmp = pe.Mat().create()
    tmp.setSizes((nrow-1, ncol))
    tmp.setUp()
    newrhs = pe.Vec().create()
    newrhs.setSizes(nrow-1)
    newrhs.setUp()

    for i in range(1,nrow):
        col, vals = mymat.getRow(i)
        tmp.setValues(i-1, col, vals)
        newrhs.setValue(i-1, rhs.array[i])
    tmp.assemble()
    newrhs.assemble()
    ns = nsvec(ncol)
    #if (ns.test(tmp)):
    #    print("is a nullspace of the matrix")
    mymat.destroy()
    rhs.destroy()
    tmp.setNullSpace(ns)
    mymat = tmp
    rhs = newrhs


ksp = pe.KSP()
ksp.create(comm = pe.COMM_WORLD)

if usenormal:
    ata = pe.Mat().createNormal(mymat)
else:
    ata = mymat.transposeMatMult(mymat, None)

ksp.setType(pe.KSP.Type.LSQR)
ksp.setOperators(mymat, ata)

try:

  ksp.setFromOptions()
  ksp.setInitialGuessNonzero(True)
 
  pc = ksp.getPC()

  # for LU only, stability
  if (pc.getType() == pe.PC.Type.LU):
      pc.setFactorShift(pe.Mat.FactorShiftType.NONZERO) 
  ksp.solve(rhs, guess)
  #print("for PC ", pc.getType(), " : diverged ", ksp.is_diverged, "with numeric reason ", ksp.reason)
except Exception as e:
  for i, elem in enumerate(sys.argv):
    if elem == "-pc_type":
      print("error occured for PC", sys.argv[i+1])
      break
    
